/**
 * Opens a section
 *
 * @param {AccordionSectionList} sectionList The sections list of the accordion instance.
 * @param {AccordionSection} sectionToOpen The section to open
 * @return {boolean} true if the open sections changed, false otherwise.
 * @private
 */
function openSection(sectionList, sectionToOpen) {
    let sectionChanged = false;

    if (sectionToOpen) {
        const sections = sectionList.sections;
        sections.forEach(section => {
            if (section !== sectionToOpen) {
                sectionChanged = sectionChanged || section.isOpen();
                section.close();
            }
        });
        sectionChanged = sectionChanged || !sectionToOpen.isOpen();
        sectionToOpen.open();
    }

    return sectionChanged;
}

export class SingleOpenSectionStrategy {
    /**
     *
     * @param {AccordionSectionList} sectionList the sections handled by this strategy.
     */
    constructor(sectionList) {
        this.privateSectionList = sectionList;
    }

    openFirstSection() {
        const sections = this.privateSectionList.sections;

        if (sections.length) {
            return openSection(this.privateSectionList, sections[0]);
        }

        return false;
    }

    openSectionByName(name) {
        const normalizedName =
            Array.isArray(name) && name.length > 0 ? name[0] : name;
        const sectionToOpen = this.privateSectionList.getSectionByName(
            normalizedName
        );

        return openSection(this.privateSectionList, sectionToOpen);
    }

    handleSectionSelect(sectionId) {
        const selectedSection = this.privateSectionList.getSectionById(
            sectionId
        );

        return openSection(this.privateSectionList, selectedSection);
    }

    handleSectionWillDeregister(sectionId) {
        const sectionToDelete = this.privateSectionList.getSectionById(
            sectionId
        );

        if (sectionToDelete.isOpen()) {
            const sectionToOpen = this.privateSectionList.getNextSectionTo(
                sectionId
            );

            if (sectionToOpen && sectionToOpen !== sectionToDelete) {
                sectionToOpen.open();
            }

            return true;
        }

        return false;
    }
}
