Use the `lightning/messageService` module to communicate across the DOM between Aura components, Visualforce pages, and Lightning web components. Lightning Message Service also enables communication between a component in a main Lightning page with one in a pop-out utility bar.

A Lightning web component uses a Lightning Message Channel to access the Lightning Message Service API. Reference a Lightning Message Channel with the scoped module `@salesforce/messageChannel`. 

A message is a serializable JSON object. Examples of data that you can pass in a message include strings, numbers, booleans, and objects. A message cannot contain functions and symbols. 

This example shows how to subscribe a Visualforce page to a message channel on which a Lightning web component publishes. It assumes that you created a message channel called `SampleMessageChannel__c` using the Metadata API. 

Create a Lightning web component called publisherComponent that publishes on `SampleMessageChannel__c`, the API name of the message channel. The component imports methods to publish on a message channel from the `lightning/messageService` module. This module accesses the Lightning Message Service API. Next, import the message channel from `@salesforce/messageChannel/SampleMessageChannel__c` and assign it to the identifier `SAMPLEMC`.

The `handleClick()` method holds the data to be published on the message channel. Here, the data is a `recordId` with the value "some string" and `recordData`, whose value is the key-value pair value: "some data". Publish the data by calling the `publish()` method imported from the `lightning/messageService` module.


```javascript
// publisherComponent.js
import { LightningElement, wire } from 'lwc';
import { publish, MessageContext } from 'lightning/messageService';
import SAMPLEMC from "@salesforce/messageChannel/SampleMessageChannel__c";

export default class PublisherComponent extends LightningElement {
    @wire(MessageContext)
    messageContext;

    handleClick() {
        const message = {
            recordId: "some string",
            recordData: {
                value: "some data"
            }
        };
        publish(this.messageContext, SAMPLEMC, message);
    }
}
```

In your component’s HTML template file, include a Publish button that calls the `handleClick()` method. Clicking the button publishes the record data to `SampleMessageChannel__c`. The subscribing Visualforce page, Aura component, or Lightning web component then receives that data.

```html
<-- publisherComponent.html -->
<template>
    <lightning-card title="MyLwcPublisher" icon-name="custom:custom14">
        <div class="slds-m-around_medium">
            <p>MessageChannel: SampleMessageChannel</p>
            <br>
            <lightning-button label="Publish" onclick={handleClick}></lightning-button>
        </div>
    </lightning-card>
</template>
```



#### See Also
[Lightning Message Channel (Beta)](docs/component-library/documentation/lwc/lwc.use_message_channel)
