---
examples:
 - name: basic
   label: Stateful Button Icons
   description: Stateful button icons can be toggled between states. They support multiple sizes and can be disabled.
 - name: variants
   label: Stateful Button Icons with Border Variants
   description: Button icons using the two types of border variants.
---
A `lightning-button-icon-stateful` component represents an icon-only button
element that toggles between two states. For example, you can use this
component for capturing a customer's feedback on a blog post (like or
dislike). Clicking the button triggers the handler set for `onclick` and changes the state of the icon using the `selected` attribute.

This example creates a Like button that toggles between two states. The Like button is selected by default. The button's state is stored in the `selected` attribute.

Selecting the Dislike button also toggles the state on the Like button and deselects it.

```html
<template>
    <lightning-button-icon-stateful
        icon-name="utility:like"
        selected={liked}
        alternative-text="Like"
        onclick={handleToggle}>
    </lightning-button-icon-stateful>
</template>
```

Handle the `click` event in your JavaScript code.

```javascript
import { LightningElement } from 'lwc';

export default class MyComponentName extends LightningElement {
    liked = true;

    handleToggle() {
        this.liked = !this.liked;
    }
}
```

#### Component Styling

`lightning-button-icon-stateful` implements the 
[button icons](https://www.lightningdesignsystem.com/components/button-icons/) blueprint in the Salesforce Lightning Design System (SLDS).

You can use a combination of the `variant` and `size` attributes to
customize the button and icon styles. 

##### Icons

Use the `icon-name` attribute to add a utility icon to the button.

The SLDS utility icon category offers nearly 200 utility
icons that can be used in `lightning-button-icon-stateful`. Although the
SLDS provides several categories of icons, only the utility
category can be used with this component.

Visit [utility icons](https://lightningdesignsystem.com/icons/#utility) in
the SLDS to view the utility icons.

When applying SLDS classes or icons, check that they are
available in the SLDS release tied to your org. The latest
SLDS resources become available only when the new release is available in your org.

##### Variants

Use the `variant` attribute with one of these values to apply styling.

* `border` is the default variant, an icon in a transparent container with a border
* `border-filled` displays an icon in a filled container with a border
* `border-inverse` displays an icon in a transparent container with a border, useful for dark backgrounds

##### Sizes

Adjust the button and icon sizes using the `size` attribute with one of these values.

* `medium` is the default size, which creates a 32px by 32px button enclosing a 14px by 14px icon.
* `small` creates a 24px by 24px button enclosing a 14px by 14px icon
* `x-small` creates a 20px by 20px button enclosing a 12px by 12px icon
* `xx-small` creates a 16px by 16px button enclosing a 8px by 8px icon

##### Customize Component Styling

To apply additional styling, use the SLDS [utility classes](https://www.lightningdesignsystem.com/utilities/alignment) with the `class` attribute.

This example adds a margin to the left of the second button using an SLDS class.

```html
<lightning-button-icon-stateful icon-name="utility:like" 
    selected={likeState}
    onclick={handleLikeButtonClick}
    alternative-text="Like"></lightning-button-icon-stateful>
<lightning-button-icon-stateful icon-name="utility:answer"
    selected={answerState} 
    onclick={handleAnswerButtonClick} 
    alternative-text="Answer"
    class="slds-m-left_xx-small"></lightning-button-icon-stateful>
```

To change the background color on a button, define a custom class using the `class` attribute. This example applies a purple background on a button using the `border-inverse` variant.

```html
 <lightning-button-icon-stateful icon-name="utility:like" 
            selected={likeState}
            onclick={handleLikeButtonClick}
            alternative-text="Like"
            variant="border-inverse"
            class="my-brand"></lightning-button-icon-stateful>
```

Specify the background color on the button using the `--sds-c-button-color-background` custom property.

```css
.my-brand {
    --sds-c-button-color-background: purple;
}
```

For more information, see [Style Components Using Lightning Design System Styling Hooks](docs/component-library/documentation/lwc/lwc.create_components_css_custom_properties) in the Lightning Web Components Developer Guide.

#### Usage Considerations

Icons are not available in Lightning Out, but they are available in Lightning Components for Visualforce and other experiences.