import { normalizeString } from 'lightning/utilsPrivate';
import { WindowManager } from './util';

// TODO: Remove, not currently in use.
const ALIGN_REGEX = /^(left|right|center)\s(top|bottom|center)$/;

export const Direction = {
    Center: 'center',
    Middle: 'middle',
    Right: 'right',
    Left: 'left',
    Bottom: 'bottom',
    Top: 'top',
    Default: 'default',
};

const VerticalMap = {
    top: Direction.Top,
    bottom: Direction.Bottom,
    center: Direction.Middle,
};

const HorizontalMap = {
    left: Direction.Left,
    right: Direction.Right,
    center: Direction.Center,
};

const FlipMap = {
    left: Direction.Right,
    right: Direction.Left,
    top: Direction.Bottom,
    bottom: Direction.Top,
    center: Direction.Center,
    default: Direction.Right,
};

function getWindowSize() {
    return {
        width:
            WindowManager.window.innerWidth || document.body.clientWidth || 0,
        height:
            WindowManager.window.innerHeight || document.body.clientHeight || 0,
    };
}

export function normalizeDirection(direction, defaultValue) {
    return normalizeString(direction, {
        fallbackValue: defaultValue || Direction.Default,
        validValues: [
            Direction.Center,
            Direction.Right,
            Direction.Left,
            Direction.Bottom,
            Direction.Top,
            Direction.Middle,
            Direction.Default,
        ],
    });
}

export function mapToHorizontal(value) {
    value = normalizeDirection(value, Direction.Left);
    return HorizontalMap[value];
}

export function mapToVertical(value) {
    value = normalizeDirection(value, Direction.Left);
    return VerticalMap[value];
}

export function flipDirection(value) {
    value = normalizeDirection(value, Direction.Left);
    return FlipMap[value];
}

// TODO: Remove, not currently in use.
export function isValidDirection(value) {
    return value && value.match(ALIGN_REGEX);
}

export function checkFlipPossibility(element, target, leftAsBoundary) {
    const viewPort = getWindowSize();
    const elemRect = element.getBoundingClientRect();
    const referenceElemRect = target.getBoundingClientRect();
    const height =
        typeof elemRect.height !== 'undefined'
            ? elemRect.height
            : elemRect.bottom - elemRect.top;
    const width =
        typeof elemRect.width !== 'undefined'
            ? elemRect.width
            : elemRect.right - elemRect.left;

    // TODO: We'll need to revisit the leftAsBoundary config property. Either we'll need a better
    // name to cover the RTL language cases and maybe open up the possibility of bounding the
    // element to the target in both the horizontal and vertical directions.

    // The boundary shrinks the available area to the edge of the target rather than the viewport.
    let rightAsBoundary = false;
    if (document.dir === 'rtl') {
        rightAsBoundary = leftAsBoundary;
        leftAsBoundary = false;
    }

    const aboveSpace = referenceElemRect.top - height;
    const belowSpace = viewPort.height - referenceElemRect.bottom - height;

    const hasSpaceAbove =
        aboveSpace >= 0 || (belowSpace < 0 && aboveSpace > belowSpace);
    const hasSpaceBelow =
        belowSpace >= 0 || (aboveSpace < 0 && belowSpace > aboveSpace);

    // Assuming left alignment is specified this tests if:
    // - there's room to accommodate the element with right alignment
    // - there's not enough room to accommodate the element with left alignment
    const shouldAlignToRight =
        referenceElemRect.right >= width &&
        referenceElemRect.left + width >
            (rightAsBoundary ? referenceElemRect.right : viewPort.width);

    // Assuming right alignment is specified this tests if:
    // - there's room to accommodate the element with left alignment
    // - there's not enough room to accommodate the element with right alignment
    const shouldAlignToLeft =
        referenceElemRect.left + width <= viewPort.width &&
        referenceElemRect.right - width <
            (leftAsBoundary ? referenceElemRect.left : 0);

    // Assuming center alignment, does the viewport have space to fit half of the element around
    // the target?
    const centerOverflow = {
        left: referenceElemRect.left - width * 0.5 < 0,
        right: referenceElemRect.right + width * 0.5 > viewPort.width,
        top: referenceElemRect.top - height * 0.5 < 0,
        bottom: referenceElemRect.bottom + height * 0.5 > viewPort.height,
    };

    return {
        shouldAlignToLeft,
        shouldAlignToRight,
        hasSpaceAbove,
        hasSpaceBelow,
        centerOverflow,
    };
}
